%% Call graphs handling module for Sloth.
%% Copyright (C) 2006, Babel Group
%%
%% Written by: Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(cgraph, [build_cgraph/2, cgraph_get_visited/2, cgraph_draw/1]).

:- use_module(curryast, [get_calls/2]).

:- use_module(library(lists), [union/3, difference/3, list_concat/2]).

%% type: call_graph 
%% TODO: Use Ciao types.
%% cg_leaf(F, [])
%% cgraph(F, [Cgraphs])
%% cg_fix([Fs], [Cgraphs])

%% build_cgraph(Function, CallTree)
build_cgraph(F, CG) :-
	build_cgraph_aux(F, [], CG).

%% build_cgraph_aux(+Function, +Visited, -CallG).
build_cgraph_aux(F, V, cg_fix([F|FL], [])) :-
	member(F, V),
	!,
	append(FL, [F|_], V).

build_cgraph_aux(F, _V, cg_leaf(F, [])) :-
	get_calls(F, []),
	!.

build_cgraph_aux(F, V, Res) :-
	get_calls(F, FL),
	get_cgraph_all(FL, [F|V], CGL),
	merge_fixpoints(F, CGL, Res).
	    
get_cgraph_all([], _V, []).
get_cgraph_all([F|FL], V, [C|CG]) :-
	build_cgraph_aux(F, V, C),
	!,
	get_cgraph_all(FL, V, CG).

%% Fixpoint merging is complicated. I haven't come with a better way
%% to do it that a case by case strategy.
merge_fixpoints(F, CGList, Res) :-
	findall(
		   cg_fix(FL, Acc), 
		   member(cg_fix(FL, Acc), CGList),
		   LFix
	       ),
	difference(CGList, LFix, LNoFix),
	(
	    %% We've got no cg_fix, no merging needed.
	    LFix = [] ->
	    Res = cgraph(F, CGList)
	;
	    %% We have some cg_fix.
	    %% Either we've to propagate it upwards or it's a terminal one.
	    merge_cg_fix(F, LFix, Merged),
	    (
		Merged = cgraph(F, Calls1) ->
		append(Calls1, LNoFix, ResCalls),
		Res = cgraph(F, ResCalls)
	    ;
		Merged = cg_fix(FL, Calls1) ->
		append(Calls1, LNoFix, ResCalls),
		Res = cg_fix(FL, ResCalls)
	    )
	).
%%	format("DEBUG:~p~n", [L]).

merge_cg_fix(F, [cg_fix(LF, Calls)], cg_fix(LF, Calls)) :-
	member(F, LF),
	!.

merge_cg_fix(F, [cg_fix(LF, Calls)], cgraph(F, [cg_fix(LF, Calls)])) :-
	!.

merge_cg_fix(F, [Cg|RR], CgR) :-
	merge_cg_fix(F, RR, MergedRR),
	merge_cg_fix(F, [Cg], MergedCg),
	(
	    MergedRR = cgraph(F, Calls1) ->
	    (
		MergedCg = cgraph(F2, Calls2) ->
		union(Calls1, Calls2, RCalls),
		CgR = cgraph(F, RCalls)
	    ;
		%% MergedCg = cg_fix && MergedRR = cgraph
	        MergedCg = cg_fix(FL, CallsFix),
		union(CallsFix, Calls1, RCalls),
		CgR = cg_fix(FL, RCalls)
	    )
	;
	    MergedRR = cg_fix(FL, Calls1),
	    (
		MergedCg = cgraph(F2, Calls2) ->
		union(Calls1, Calls2, RCalls),
		CgR = cg_fix(FL, RCalls)
	    ;
		%% MergedCg = cg_fix && MergedRR = cg_fix
	        MergedCg = cg_fix(FL2, CallsFix),
		union(CallsFix, Calls1, RCalls),
		union(FL, FL2, FRR),
		CgR = cg_fix(FRR, RCalls)
	    )
	).


%% Get the list of functions in a call graph.
cgraph_get_visited(cg_leaf(F, []), [F]) :- !.
cgraph_get_visited(cg_fix(FL, CG),  Nodes) :-
	!,
	cgraph_get_visited_list(CG, CNodes),
	append(FL, CNodes, Nodes).
cgraph_get_visited(cgraph(F, LL), [F|Nodes]) :-
	!,
	cgraph_get_visited_list(LL, NN),
	list_concat(NN, Nodes).

cgraph_get_visited_list([], []).
cgraph_get_visited_list([F|FF], [R|RR]) :-
	cgraph_get_visited(F, R),
	!,
	cgraph_get_visited_list(FF, RR).

%% Drawing 
cgraph_draw(CGraph) :-
	cgraph_draw_aux(CGraph, 0).
cgraph_draw_aux(cg_leaf(F, []), N) :-
	format("~*c->~p [Leaf] ~n", [N, 0' , F]).
cgraph_draw_aux(cg_fix(F, CallL), N) :-
	format("~*c->~p [Cycle] ~n", [N, 0' , F]),
	N4 is N + 4,
	cgraph_draw_list(CallL, N4).

cgraph_draw_aux(cgraph(F, CallL), N) :-
	format("~*c->~p ~n", [N, 0' , F]),
	N4 is N + 4,
	cgraph_draw_list(CallL, N4).

cgraph_draw_list([], _).
cgraph_draw_list([L|LL], N) :-
	cgraph_draw_aux(L, N),
	!,
	cgraph_draw_list(LL, N).
