:- module(debug, [dump_abstree/2, dump_rules/2, 
	          dump_symtab/2,  dump_deftrees/2, 
		  pretty_list_write/1, pretty_write/1, diff_lists/3,
	          debug_write/1, debug_write_string/1,
		  debug_pass/1] ).

:- use_module(library(lists),      [difference/3, append/3]).
:- use_module(library(strings),    [write_string/2] ).
:- use_module(library(hiordlib),   [map/3] ).
:- use_module(library(aggregates), [findall/3]).

% Library modules.
:- use_module(symtab,  [get_real_rules/2, get_object_property/3] ).
:- use_module(cmdopts, [cmd_options/2]).
:- use_module(deftree, [call_defTree/2]).

%% For a better pretty printing, we include operators.
:- include(ast_ops).

%% Prints the relevant info at pass' exit.
debug_pass(When) :-
	cmd_options('-tp', _),
	!,
	statistics(runtime, [_,Time1]),
	format("~20s: ~f~n", [When, Time1]).
debug_pass(_).

% Debug write
debug_write(E) :-
	write(user_error, E).

debug_write(pos(L, C)) :-
	write(user_error, 'Line: '),
	write(user_error, L),
	write(user_error, ' Col: '),
	write(user_error, C).

debug_write_string(S) :-
	write_string(user_error, S).

dump_rules(_Module, Rules) :-
	findall(rule(N, R), get_real_rules(N, R), Rules),
	!.

dump_symtab(Module, Entries) :-
	findall(entry(N, P, V), 
	    (
		get_object_property(N, P, V),
		N = qid(Module, _)
	    ), Entries),
	!.	

dump_deftrees(_M, DefTrees) :-
	findall(deftree(N, Dt), (
				    get_real_rules(N, _R),
				    call_defTree(N, Dt)
				), 
				DefTrees),
	!.
				    

pretty_write(rule(N,R)) :-
	pretty_write(N),
	debug_write(': '),
	debug_write('\n'),
	prettyvars(R),
	pretty_list_write(R),
	debug_write('\n').

pretty_write(entry(N, P, V)) :-
	pretty_write(N),
	debug_write('.'),
	debug_write(P),
	debug_write(': '),
	prettyvars(V),
	debug_write(V),
	debug_write('\n').

pretty_write(qid(Module, Name)) :-
	debug_write(Module),
	debug_write('.'),
	debug_write(Name).

pretty_write(plus(E)) :-
 	debug_write('+ '),
	pretty_write(E).
%%	write('\n').

pretty_write(minus(E)) :-
 	debug_write('- '),
	pretty_write(E).
%%	write('\n').

pretty_write(deftree(N, D)) :-
	debug_write('\nDeftree for '),
	debug_write(N),
	debug_write('\n-----------------\n'),
	debug_write_deftree(D, 0).

pretty_write(A := B) :-
	pretty_write(A),
	debug_write(' := ' ),
	pretty_write(B),
	debug_write('\n').

pretty_write(E@Params) :-
	pretty_write(E),
	debug_write('@['),
	pretty_list_write_deco(Params),
	debug_write(']').

pretty_write(const(_, C)) :-
	pretty_write(C).

pretty_write(E) :-
	!,
	debug_write(E).

%% This is very wrong, we should fix this.
indent(0).
indent(N) :- 
	debug_write(' '),
	N1 is N - 1,
	indent(N1).

debug_write_deftree(fallo-[], N) :-
	indent(N),
	debug_write('Fallo en deftree\n'),
	!.

debug_write_deftree(regla-[Rule], N) :-
	indent(N),
	debug_write('regla -> '),
	debug_write('\n'),
	N1 is N + 2,
	indent(N1),
	pretty_write(Rule),
%%	portray_clause(user_error, (Rule)),
%%	debug_write('\n'),
	!.

debug_write_deftree((Cons, Dt), N) :-
	indent(N),
	debug_write(Cons),
	debug_write(' =>\n'),
	N1 is N + 2,
	debug_write_deftree(Dt, N1),
	!.

debug_write_deftree(branch(Pivot, SubTrees)-_Ns, N) :-
	indent(N),
	debug_write('-- Branch by '),
	debug_write(Pivot),
	length(SubTrees, LN),
	debug_write(': '),
	debug_write(LN),
	debug_write('\n'),
	N1 is N + 4,
	debug_write_deftrees(SubTrees, N1),
	indent(N),
	debug_write('-- End Branch '),
	!.

debug_write_deftree(or(SubTrees)-_Ns, N) :-
	indent(N),
	debug_write('--- Or branch:     '),
	length(SubTrees, LN),
	debug_write(LN),
	debug_write('\n'),
	N1 is N + 4,
	debug_write_deftrees(SubTrees, N1),
	%debug_write_deftree(NS),
	debug_write('\n'),
	indent(N),
	debug_write('--- End Or branch: '),
	debug_write(LN),
	debug_write('\n'),
	!.

debug_write_deftrees([], _).
debug_write_deftrees([D|DL], N) :-
	debug_write_deftree(D, N),
	debug_write('\n'),
	debug_write_deftrees(DL, N).

pretty_list_write([]).
pretty_list_write([E|ES]) :-
	pretty_write(E),
	pretty_list_write(ES).


pretty_list_write_deco([]).
pretty_list_write_deco([E|ES]) :-
	pretty_write(E),
	(
	    \+ ES = [],
	    debug_write(', '),
	    pretty_list_write_deco(ES)
	;
	    true
	).


% Debug function, writes symbol table to file.
dump_abstree(Module, Abstree) :-
	dump_rules(Module, Rules),
	dump_symtab(Module, St),
	append(Rules, St, Abstree),
	!.


%%% This predicate generates a unified diff style list difference.

mark(E, M, R) :-
	R =.. [M|[E]].

diff_lists(L1, L2, LR) :-
	difference(L1, L2, Minus),
	difference(L2, L1, Plus),
	map(Minus, mark(minus), Min),
	map(Plus,  mark(plus),  Plu),
	append(Min, Plu, LR).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

