%% Sloth's deftree generation.
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey
%%
%% Modified by:
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(deftree,	 [defTree/3, call_defTree/2] ).

% Module use.

:- use_module(symtab,      [get_rules/2, get_object_property/3, object_arity/2] ).
:- use_module(const_expr,  [is_constant/1] ).
:- use_module(curryast,    [split_rule/3, param_at/3, get_function/2, get_params/2] ).

%

:- use_module(library(lists),      [append/3, length/2, select/3, reverse/2,
	                            difference/3]).
:- use_module(library(aggregates), [findall/3]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Aux functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Esto es un takeN
comienzos(0, []):-!.
comienzos(N, [[N] | R]):-
	N > 0,
	N1 is N-1,
	comienzos(N1, R).

%% Pertenencia a lista con backtracking.
in(X,[X|_]).
in(X,[_|Xs]) :- 
	in(X,Xs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pivot calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mapAtPos([],_,[]).
mapAtPos([R|Rs], Pos, [T|Ts]) :-
	split_rule(R, Head, _),
	param_at(Head, Pos, T),
	mapAtPos(Rs, Pos, Ts).

uniform(Reglas,Pos) :-
	mapAtPos(Reglas, Pos, Terms),
	\+ in(param(_), Terms),
	\+ in(freevar(_), Terms).

% Select the pivot by getting the first position that has not free variables 
% in it. Be warned that this select all the possible pivots on backtracking.
pivote(Reglas, Poss, Piv):-
	in(Piv, Poss),
	uniform(Reglas, Piv).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Split3 calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Aux predicates for traduce.
posiciones(0,[]):-!.
posiciones(N,[N|Ns]):-
	N1 is N-1,
	posiciones(N1,Ns).

prepend(_,[],[]).
prepend(Pos,[Q|Qs],[PQ|PQs]):-
	append(Pos,[Q],PQ),
	prepend(Pos,Qs,PQs).

%%
distrib_rules(Reglas, Con, Pos, R):- % Parser -> Parser
	in(R, Reglas),
	split_rule(R, Head, _),  
	param_at(Head, Pos, Term),
	get_function(Term, Con_aux),
	Con = Con_aux. 

%% 
% traduce([], _, _, []).
% traduce([Con-Arity|Casos], Reglas, Pos, [(Con, ReglasCon, NPosPos)|Subs]) :-
% 	findall(
% 		   R, 
% 		   distrib_rules(Reglas, Con, Pos, R), 
% 		   ReglasCon
% 	       ),
% 	posiciones(Arity, NPos),
% 	prepend(Pos,NPos,NPosPos),
% 	traduce(Casos,Reglas,Pos,Subs).

%% Map cons and rules to (cons, rules)
%% Warning, we have to follow rule order here, if not bad things happen.
assign_cons(_, [], _, []).
assign_cons(Cons, [Rule|RL], Pos, [(Con, ReglasCon, NPosPos)|Subs]) :-
	Reglas = [Rule|RL],
	%% Get the rule cons in the demanded position.
	split_rule(Rule, Head, _),
	param_at(Head, Pos, Term),
	get_function(Term, Con),
	select(Con-Arity, Cons, RestC),
	findall(
		   R,
		   distrib_rules(Reglas, Con, Pos, R),
		   ReglasCon
	       ),
	posiciones(Arity, NPos),
	prepend(Pos,NPos,NPosPos),
	difference(Reglas, ReglasCon, RestR),
	assign_cons(RestC, RestR, Pos, Subs).

%
lista_const([T|Ts], Pos, Ac, Nums) :- 
	!,
	lista_const(T,  Pos, Ac, N1),
	lista_const(Ts, Pos, N1, Nums).

lista_const(Const, _Pos, Ac, [Const-0|Ac]) :-
	is_constant(Const),
	\+ member(Const-0, Ac),
	!.

lista_const(Rule, Pos, Ac, Ac1):-
	split_rule(Rule, Head, _),
	!,
	lista_const(Head, Pos, Ac, Ac1).

lista_const(Term, Pos, Ac, Nums):-
	param_at(Term, Pos, Arg),
	!,
	lista_const(Arg, Pos, Ac, Nums).

lista_const(_, _, Ac, Ac).

%% Finds all possible cases for a parameter. 
%% This effectively gives us the number of childs of the deftree.

casos(_, _, Casos, Con) :-
	get_object_property(Con, constructs, Type),
	!,
	findall(Cons-Arity,
	    (
		get_object_property(Cons, constructs, Type),
		object_arity(Cons, Arity)
	    ), 
	    Casos).

%% When we split by a constant, construct the list 
%% of them for the tree.
casos(Reglas, Pos, NumsReglas2, X) :-
	is_constant(X),
	!,
	lista_const(Reglas, Pos, [], NumsReglas),
	reverse(NumsReglas, NumsReglas2).


%% split3(Reglas, Pos, SubCasos).
%% split3 returns a list of subtrees ignoring the current pivot.

split3([Regla| Reglas], Pos, Subs) :-
	split_rule(Regla, Head, _),
	param_at(Head, Pos, PivotTerm),
	get_function(PivotTerm, Pivot),
        casos([Regla|Reglas], Pos, Casos, Pivot),
	assign_cons(Casos, [Regla|Reglas], Pos, Subs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Split3 calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% defTree(Reglas, Posiciones, Arbol)

defTree([],_,fallo-[]).

%% The main case is a branch or rule..
defTree(Reglas, Poss, branch(Piv, Subs)-Ns):-
	pivote(Reglas, Poss, Piv),
	Reglas = [_|_],
	!,
	fake = Ns,  
	split3(Reglas, Piv, Casos), 
	trans(Casos, Piv, Poss, SubAux),
	mapDefTree(SubAux, Subs).

defTree([Regla],Poss,regla-[N]):-
	\+(pivote([Regla],Poss,_)),
	Regla = N.

%% Beware parallel trees maybe broken by now.
defTree(Reglas,Poss,or(DefTrees)-Ns):-
	\+ pivote(Reglas,Poss,_),
	Reglas=[_,_|_],
	Reglas = Ns,
	map_regla_defTree(Reglas,Poss,DefTrees).

map_regla_defTree([],_,[]).
map_regla_defTree([Regla|Reglas],Poss,[DefTree|DefTrees]) :-
	defTree([Regla], Poss, DefTree),
	map_regla_defTree(Reglas, Poss, DefTrees).


mapDefTree([],[]).
mapDefTree([(Con,Reglas,Poss)|Ss],[(Con,T)|Ts]) :-
	defTree(Reglas, Poss, T),
	mapDefTree(Ss, Ts).

trans([],_,_,[]).
trans([(Con,Reglas,NPoss)|Casos],Piv,Poss,[(Con,Reglas,NNPoss)|NCasos]) :-
	append(Antes, [Piv|Despues], Poss),
	append(NPoss, Despues, Aux),
	append(Antes, Aux, NNPoss),
	trans(Casos, Piv, Poss, NCasos).

%%%%%%%%%%%%%%%%%%%%% Arboles Paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
call_defTree(Pred, Arbol):-
	conjuntos_reglas(Pred,Conjuntos_reglas),
	length(Conjuntos_reglas, L),
	L > 1,
	!,
	%% Rule overlapping as in coin = 1; coin = 2, this is
        %% different from paralell or trees, because here the rules
        %% are truly non-deterministic.
	object_arity(Pred, Arity),
	comienzos(Arity, Comienzos),
	findall(A,(
		      member(Conjunto,Conjuntos_reglas),
		      Conjunto = Reglas,
		      defTree(Reglas,Comienzos,A)
		  ),
		  Arboles),
	get_rules(Pred, Ns),
	reverse(Arboles, RArboles),
	Arbol = or(RArboles)-Ns.

call_defTree(Pred,Arbol):-
	object_arity(Pred, Arity),
	!,
	comienzos(Arity, Comienzos),
	get_rules(Pred, Reglas),
	defTree(Reglas,Comienzos,Arbol).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cociente_reglas(Pred,Cociente):-
	% Encuentra todas las Id de regla para pred
        get_rules(Pred, Rules),
	% Las mete a todas las reglas en un lista.
        findall([Id],member(Id,Rules),Cociente0),
	cociente_reglas_aux(Pred,Cociente0,Cociente).

cociente_reglas_aux(_Pred, [], []) :- !.
cociente_reglas_aux(_Pred, [Conj], [Conj]) :- !.
cociente_reglas_aux(Pred, Cociente_ini, Cociente) :-
	% Coge con backtracking una lista de regla
	select(CR1, Cociente_ini, Cociente_ini1),
	select(CR2, Cociente_ini1, Cociente_ini2),
	CR1 = [R1|_],
	CR2 = [R2|_],
	split_rule(R1, H1, _), get_params(H1, P1),
	split_rule(R2, H2, _), get_params(H2, P2),
	%% Params are equivalent ie : f x = 1 ; f y = 2, so we have an or tree
%	P1 = P2,
        copy_term(P1, PP1), copy_term(P2, PP2),
        eq_params(PP1, PP2),
	!,
	append(CR1, CR2, Conj1_2),
	append([Conj1_2], Cociente_ini2, Cociente1),
	cociente_reglas_aux(Pred, Cociente1, Cociente).

cociente_reglas_aux(_,Cociente,Cociente).

%% FIXME: We should get this to its proper place.
:- include(ast_ops).
eq_params([], []).
eq_params([A|AL], [B|BL]) :-
	eq_params(A, B),
	eq_params(AL, BL).
eq_params(param(A), A).
eq_params(A, param(A)).
eq_params(A@A1, B@B1) :-
	eq_params(A, B),
	eq_params(A1, B1).
eq_params(A, A).	
%%%%%%% Arboles paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% uno_de_cada(Conjuntos,Conjunto) toma una lista de listas y devuelve con 
% backtracking conjuntos con un elemento de cada uno de los Conjuntos

uno_de_cada([],[]).
uno_de_cada([Lista|Listas],[Elemento|Elementos]):-
	member(Elemento,Lista),uno_de_cada(Listas,Elementos).

% conjuntos_reglas(Predicado,Conjuntos) toma el nombre de un predicado y devuel-
% ve todos los conjuntos de reglas posibles con los que se puede formar un
% arbol determinista. Las reglas se indican por su numero.

conjuntos_reglas(Pred, Conjs):-
	cociente_reglas(Pred,Cociente),
	findall(C,uno_de_cada(Cociente,C),ConjsAux),
	quitar_repetidos(ConjsAux,[],_,ConjsRev),  
	reverse(ConjsRev, Conjs).


quitar_repetidos_lista([],Ac,Ac,[]).
quitar_repetidos_lista([R|Rs],AcIn,AcOut,[R|Rs2]):-
	\+member(R,AcIn),!,
	quitar_repetidos_lista(Rs,[R|AcIn],AcOut,Rs2).
quitar_repetidos_lista([_R|Rs],AcIn,AcOut,Rs2):-
	quitar_repetidos_lista(Rs,AcIn,AcOut,Rs2).

quitar_repetidos([],Ac,Ac,[]).
quitar_repetidos([Lista|Listas],AcIn,AcOut,[Lista2|Listas2]):-
	quitar_repetidos_lista(Lista,AcIn,AcIn2,Lista2),
	quitar_repetidos(Listas,AcIn2,AcOut,Listas2).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

