%% Sloth's export/import interface handling.
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by:
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(exports, [load_exports/2, fix_exports/1]).

:- use_module(library(lists),      [append/3] ).
:- use_module(library(aggregates), [findall/3] ).
:- use_module(library(hiordlib),   [map/3]).

%%
:- use_module(types,             [type_name/2]).
:- use_module(scope,             [open_scope/1, close_named_scope/1,
	                          lookup_symbol/2, insert_symbol/2]).
:- use_module(symtab,            [get_module_functions/2, get_module_types/2,
	                          add_rule/2, set_object_property/3, new_type/1,
				  replace_object_property/3, is_type/1,
				  get_object_property/3, add_constructor/2] ).
:- use_module(curryast,          [function_name/2]).


% We should have here:
% * Tools for loading the exports into the symbol table.
% * Tools for converting the various internal interface representation to Prolog and viceversa.
% * Import checking

%% This loads all the imports from the module. Imports can be in the following way:
%% - id(Name)
%% - type(Name, [Cons1..Consn])

load_exports(Module, Imports) :-
	use_module(sloth(Module)),
	open_scope(imported(Module)),
	findall(RIface, (
			   Module:exports(Iface),
			   check_import(Iface, Imports, RIface)
			), Exports
	       ),
	(
	    reverse(Exports, EE),
	    load_interfaces(EE),
%	    load_interfaces(Exports),
	    !
	;
	    throw(error(symtab, iface_loading(Exports)))
	).

%% This checks if we should import the id or type.
check_import(Import, [], Import) :-
	!.
check_import(id(Name, Type, Assoc), Imports, id(Name, Type, Assoc)) :-
	function_name(Name, FName),
	member(id(FName), Imports).

check_import(type(Name, Cons), Imports, type(Name, RCons)) :-
	type_name(Name, TName),
	TName = tid(_M, TN),
	member(type(TN, ImportCons), Imports),
	(
	    ImportCons = [] ->
	    RCons = ImportCons
	;
	    intersect_cons(Cons, ImportCons, RCons)
	).

%% Check that we only import the correct constructors.
intersect_cons([], _, []).
intersect_cons([(C, Type)|CS], ICons, R) :-
	function_name(C, CName),
	(
	    member(CName, ICons) ->
	    R = [(C, Type)|RR]
	;
	    R = RR
	),
	intersect_cons(CS, ICons, RR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Fill symbol table with a module's interface.
%% This also fill the current scope.

load_interfaces([]).
load_interfaces([I|IF]) :-
	load_interface(I),
	!,
	load_interfaces(IF).

load_interface(end).
load_interface(id(Symbol, Type, OpAttr)) :-
	add_rule(Symbol, builtin(imported)),
	set_object_property(Symbol, type, Type),
	(
	    OpAttr = op(Assoc, Prio) ->
	    set_object_property(Symbol, assoc, Assoc),
	    set_object_property(Symbol, priority, Prio)
	;
	    true
	),
	function_name(Symbol, UName),
	insert_symbol(qid(UName), Symbol).

load_interface(type(Type, ConsL)) :-
	new_type(Type),
	type_name(Type, TypeName),
	TypeName = tid(_Module, BareSymbol),
	insert_symbol(tid(BareSymbol), TypeName),
	load_cons(Type, ConsL).

load_cons(_, []).
load_cons(Type, [(CName, CType)|R]) :-
	add_constructor(Type, CName),
	set_object_property(CName, type, CType),
	set_object_property(CName, constructs, Type),
	load_cons(Type, R),
	function_name(CName, BareSymbol),
	insert_symbol(qid(BareSymbol), CName).


%% Fix all the due exports.
fix_exports(Module) :-
	get_object_property(Module, exports, Exports),
	(
	    Exports = all ->
	    get_module_functions(Module, FunE),
	    get_module_types(Module, TypE),
	    append(FunE, TypE, AllE),
	    replace_object_property(Module, exports, AllE)
	;
	    map(Exports, fix_export, AllE),
	    replace_object_property(Module, exports, AllE)
	).

fix_export(id(E), S) :-
	lookup_symbol(qid(E), S).
fix_export(type(T, Cons), (TS, CS)) :-
	lookup_symbol(tid(T), T1),
	type_name(TS, T1),
	is_type(TS),
	map(Cons, lookup_symbol_qid, CS).

lookup_symbol_qid(T, R) :-
	lookup_symbol(qid(T), R).
