:- module(localdefs, [desugar_ldef/1, proccess_ldef_func/2,
	              proccess_ldef_global/2, proccess_ldef_pattern/2]).

:- use_module(library(lists), [intersection/3] ).

:- use_module(curryast,   [split_rule/3, get_variables/2, map_rules/2,
	                   map_tree/3, get_function/2 ]).  

:- use_module(symtab,     [add_rule/2]).
:- use_module(combinator, [make_combinator/3, replace_comb/3] ).

% Only 1 occur.
:- include(ast_ops).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This module is based on combinators, a class of lambda expressions
%% and the Curry Prelude. For more reference you can read my master
%% Thesis, egallego.

%% Kill the patterns and the lets.
desugar_ldef(Module) :-
	map_rules(Module, proccess_ldef_func),
	map_rules(Module, proccess_ldef_global),
	map_rules(Module, proccess_ldef_pattern).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% Local definition desugaring is done in 3 steps:
%%%
%%%   * First we made al local defined functions combinators.
%%%   
%%%   * Second we globalize the new functions, this should be done in 
%%%     another pass because map_rules won't see the new expressions.
%%%
%%%   * Third, we sort by dependencies and convert all the pattern 
%%%     matching expressions to lambda expressions.

proccess_ldef_func(ldef(LDefs, Expr), ldef(NewLD, NewExpr)) :-
	map_tree(Expr, proccess_ldef_func, InnerExpr),
	make_ldef_combinators(LDefs, LDefs, InnerExpr, NewLD, NewExpr).

proccess_ldef_global(ldef(LDefs, Expr), ldef(PatternLD, Expr)) :-
	globalize_all(LDefs, PatternLD). % Kills all local fdefs.

%% In this call we should have only patterns.
proccess_ldef_pattern(ldef(LDefs, Expr), NewExpr) :-	
	map_tree(Expr, proccess_ldef_global, InnerExpr),
	lambda_sort(LDefs, NewP),
 	reduce_patterns(NewP, InnerExpr, NewExpr).

make_ldef_combinators([], NewLD, E, NewLD, E).
make_ldef_combinators([pattern(_,_)|LD], LDAll, E, LLDAll, NE) :-
	make_ldef_combinators(LD, LDAll, E, LLDAll, NE).
make_ldef_combinators([Func|L], LDAll, E, NewLDAll, NE) :-
	make_ldef_combinator(Func, LDAll, E, L1, E1),
	make_ldef_combinators(L, L1, E1, NewLDAll, NE).

%% Make combinator and substitute all the ocurrences in the inner
%% expression and in the local expressions.
make_ldef_combinator(LFunc, LDefs, Expr, NewLDefs, NExpr) :-
	make_combinator(LFunc, _NewLFunc, RInfo), 
	map_tree(Expr, replace_comb(RInfo), NExpr),
	replace_ldefs(LDefs, replace_comb(RInfo), NewLDefs).

%%
replace_ldefs([], _, []).
replace_ldefs([pattern(P, Expr)|L], F, [pattern(P, NExp)|LR]) :-
	map_tree(Expr, F, NExp),
	replace_ldefs(L, F, LR).

replace_ldefs([Func|L], F, [NewFunc|LL]) :-
	split_rule(Func, Head, Body),
	map_tree(Head, F, NHead), %% We need to replace the head of the rules.
	map_tree(Body, F, NBody),
	split_rule(NewFunc, NHead, NBody),
	replace_ldefs(L, F, LL).

%% Globalize all functions.
globalize_all([], []).
globalize_all([pattern(P, E)|L], [pattern(P, E)|LL]) :-
	globalize_all(L, LL).
globalize_all([Func|L], LL) :-
	globalize_ldef_func(Func),
	globalize_all(L, LL).

%% Generate the new global function.
globalize_ldef_func(Func) :-
	split_rule(Func, Head, _),
	get_function(Head, FName),
	add_rule(FName, Func).

%% 
reduce_patterns([], E, E).
reduce_patterns([pattern(P, E1)|L], E, NE) :-
	reduce_patterns(L, E, NE1),
	NE = lambda(P, NE1)@[E1].
	
%% Calculate the new produced vars of a lambda expr, and the needed vars.

given_vars(lambda(Pattern, _Expr), LVars) :-
	get_variables(Pattern, LVars).

needed_vars(lambda(_Pattern, Expr), LVars) :-
	get_variables(Expr, LVars).

%% A lambda-e E1 contains other E2 when given_vars(E1) intersection 
%% needed_vars(E2) \== 0
lambda_contains(E1, E2) :-
	given_vars(E1, V1),
	needed_vars(E2, V2),
	intersection(V1, V2, I),
	!,
	I \= [].

%% Sort lambdas.
lambda_sort(List, Sorted) :-
	i_sort(List, [], Sorted).

%% Easy, dirty, I know, I'm lazy
i_sort([], Acc, Acc).
i_sort([H|T], Acc, Sorted) :-
	insert(H, Acc, NAcc),
	i_sort(T,NAcc,Sorted).
   
insert(X, [Y|T], [Y|NT]) :-
	lambda_contains(X, Y),
	insert(X,T,NT).
insert(X, [Y|T], [X,Y|T] ) :-
	!.
insert(X, [], [X]).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

